/*
    DoubleSimple: DoubleRational.m
    Copyright (C) 2003-2006  Macaulay Institute

    This file is part of DoubleSimple, a series of classes for implementing
    various remedies to floating point arithmetic errors.

    DoubleSimple is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    DoubleSimple is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/

/*

Implementation of DoubleRational class. This tries to maintain a fraction as
expressed in its simplest form, with the sign on the numerator, using doubles
for the numerator and the denominator.

The value of the instance variable d of the super class DoubleSimple is 
largely ignored in this class.

*/

#import <DoubleRational.h>
#import <string.h>
#import <stdlib.h>
#import <math.h>
#import <float.h>
#ifdef INTEL
#  import <CIieeefp.h>
#else
#  import <ieeefp.h>
#endif
#import <errno.h>
#import <ctype.h>


/*Static variables */

static const double my_infinity = 1.0 / 0.0;
				// infinity clashes with declaration in math.h
				// on Cygwin platforms.

// Strange as it may appear this will not raise any exception because the 
// compiler will store +Infinity.

/* Static functions */

/*

gcd

This function returns the greatest common divisor of two integers, using 
Euclid's algorithm from Knuth (Vol II, p. 337). 

*/

static double gcd(double u, double v) {

#ifdef INTEL
  fp_pctl p = fpsetprecision(FP_PC_DBL);
#endif

  u = fabs(u);
  v = fabs(v);
  while(v > 0) {
    double r = fmod(u, v);
    u = v;
    v = r;
  }

#ifdef INTEL
  fpsetprecision(p);
#endif

  return u;
}


/*

double2rational

Convert between a double and a rational.

*/

static void double2rational(double x, double *numerat, double *denominat) {  

#ifdef INTEL
  fp_pctl p = fpsetprecision(FP_PC_DBL);
#endif
  
  switch(fpclass(x)) {
  case FP_SNAN:
  case FP_QNAN:
    fprintf(stderr, "Attempt to set DoubleRational to NaN\n");
    abort();
  case FP_NINF:
    fprintf(stderr, "Attempt to set DoubleRational to Negative Infinity\n");
    abort();
  case FP_PINF:
    fprintf(stderr, "Attempt to set DoubleRational to Positive Infinity\n");
    abort();
  case FP_NZERO:
  case FP_PZERO:
    (*numerat) = 0.0;
    (*denominat) = 1.0;
    return;
  case FP_NDENORM:
  case FP_PDENORM:
  case FP_NNORM:
  case FP_PNORM:
    (*denominat) = 1.0;
    break;

  default:
    fprintf(stderr, "Floating point class %d not recognised\n",
	    fpclass(x));
    abort();
  }

  /* If we get here, it's FP_PNORM, FP_NNORM, FP_NDENORM or FP_PDENORM */

  /* We need to multiply up by 2 until x is an integer */

  while(fmod(x, 1.0) != 0.0) {
    x *= 2.0;
    (*denominat) *= 2.0;
  }

  (*numerat) = x;

  // and these numbers are necessarily prime to each other so there's no
  // need to reduce the fraction.

#ifdef INTEL
  fpsetprecision(p);
#endif

}

/*

decimalStr2rational

Convert between a string xxx.xxx representing a decimal number and a rational.
The second argument (exp) indicates the decimal exponent.

*/

static void decimal2rational(char *str, double exp, 
			     double *numerat, double *denominat) { 

#ifdef INTEL
  fp_pctl p = fpsetprecision(FP_PC_DBL);
#endif

  char *point, *dummy, *endptr;
  double denominatorDigits = -exp;
  double greatestCommonDivisor;

  if(rint(exp) != exp) {
    fprintf(stderr, "The exponent must be an integer\n");
    abort();
  }

  dummy = strdup(str);

  point = strchr(dummy, (int)'.');

  if(point == NULL) {

    (*numerat) = strtod(dummy, &endptr);

    if(endptr == dummy) {
      fprintf(stderr, "Invalid number: %s", dummy);
      abort();
    }

  }
  else {
    char *numStr = malloc(strlen(str)*sizeof(char));
    
    (*point) = '\0';

    strcpy(numStr, dummy);
    strcat(numStr, point + 1);

    (*numerat) = strtod(numStr, &endptr);

    if(endptr == numStr) {
      fprintf(stderr, "Invalid number: %s", numStr);
      abort();
    }

    denominatorDigits += (double)strlen(point + 1);
    free(numStr);
  }

  // Consider the number of digits after the point and the exponent

  (*denominat) = 1.0;

  while(denominatorDigits > 0) {
    (*denominat) *= 10.0;
    denominatorDigits -= 1.0;
  }

  while(denominatorDigits < 0) {
    (*numerat) *= 10.0;
    denominatorDigits += 1.0;
  }

  // and reduce the fraction
  greatestCommonDivisor = gcd((*numerat), (*denominat));

  (*numerat) /= greatestCommonDivisor;
  (*denominat) /= greatestCommonDivisor;

  // make sure that we're not having -0
  if( (*numerat) == -0.0 ) (*numerat) = 0.0;

  // free the appropriate variables
  
  free(dummy);

#ifdef INTEL
  fpsetprecision(p);
#endif

} 

/* Implementation of class */

@implementation DoubleRational

/* New methods for this class */

-setNumerator: (double)numerator denominator: (double)denominator {
  double greatestCommonDivisor = gcd(numerator, denominator);

#ifdef INTEL
  fp_pctl p = fpsetprecision(FP_PC_DBL);
#endif

  num = numerator / greatestCommonDivisor;
  den = denominator / greatestCommonDivisor;

  if(den < 0.0) {
    den = -den;
    num = -num;
  }

  // avoid having -0 in the numerator
  if(num == 0.0) num = 0.0;

#ifdef INTEL
  fpsetprecision(p);
#endif

  return self;
}

-(double)getNumerator {
  return num;
}

-(double)getDenominator {
  return den;
}

/* Methods to override from DoubleSimple */

+parseValue: (char *)str {

  DoubleRational *obj = [super new];

  [obj setStr: str];

  return obj;
}

+newUp {
  DoubleRational *obj = [super newUp];

  obj->num = 0.0;
  obj->den = 1.0;

  return obj;
}

-set: (DoubleSimple *)value {
  if([value isKindOf: [DoubleRational class]]) {
    num = [(DoubleRational *)value getNumerator];
    den = [(DoubleRational *)value getDenominator];
  }
  else {
    double2rational([value get], &num, &den);
  }

  return self;
}

-cset: (double)value {
  double2rational(value, &num, &den);
  return self;
}

/*

-setStr:

A number may be represented in any of the classic floating point forms:

xxxxx
xxxxEyyy
xxx.xxx
xx.xxEyyy

Or as a rational number:

u/v

where u and v must be strings that represent integers and are representable
in double format. (e.g u = 5.03e2 is valid but u = 5.03e-2 isn't)

*/

-setStr: (char *)str {
  char *slash, *endptr, *dummyStr;

  dummyStr = strdup(str);
  num = 1.0;
  den = 1.0;

  slash = (strchr(dummyStr, (int)'/'));

  if(slash == NULL) {           // It's not u/v

    char *e, *E, *expStr;
    double exp;

    e =  strchr(dummyStr, (int)'e');
    E =  strchr(dummyStr, (int)'E');

    if(e == NULL && E == NULL) {
      decimal2rational(dummyStr, 0.0, &num, &den);
      return self;
    }

    e = (e == NULL ? E : (E == NULL ? e : (e < E ? e : E)));
    
    (*e) = '\0';
    
    expStr = strdup(e + 1);
    exp = strtod(expStr, &endptr);
    
    if(endptr == expStr) {
      fprintf(stderr, "Invalid number: %s", expStr);
      abort();
    }
    
    decimal2rational(dummyStr, exp, &num, &den);

    free(expStr);
    
  }
  else {                       // It's u/v
    
    char *denStr = strdup(slash + 1);
    double greatestCommonDivisor;
    
    (*slash) = '\0';
    
    num = strtod(dummyStr, &endptr);
    if(endptr == dummyStr) {
      fprintf(stderr, "Invalid number: %s", dummyStr);
      abort();
    }
    if(rint(num) != num) {
      fprintf(stderr, "The numerator %s is not a representable integer"
	      , dummyStr);
      abort();
    }
    
    den = strtod(denStr, &endptr);
    if(endptr == denStr) {
      fprintf(stderr, "Invalid number: %s", denStr);
      abort();
    }
    if(rint(den) != den) {
      fprintf(stderr, "The denominator %s is not a representable integer"
	      , denStr);
      abort();
    }

    // and reduce the fraction
    greatestCommonDivisor = gcd(num, den);
  
    num /= greatestCommonDivisor;
    den /= greatestCommonDivisor;

    if(den < 0.0) {
      den = -den;
      num = -num;
    }

    // avoid having -0 in the numerator
    if(num == 0.0) num = 0.0;

    free(denStr);
  }

  free(dummyStr);
  return self;
}


-(double)get {

#ifdef INTEL
  fp_pctl p = fpsetprecision(FP_PC_DBL);
#endif

  double value = num / den;

#ifdef INTEL
  fpsetprecision(p);
#endif

  return value;
}

-write: (FILE *)fp {
  double tmp = d;

  if(den == 1.0) {
    d = num;
    [super write: fp];
  }
  else {
    d = num;
    [super write: fp];
    fprintf(fp, "/");
    d = den;
    [super write: fp];
  }
  d = tmp;
  return self;
}

-write: (FILE *)fp sf: (int)sigfig {
  double tmp = d;
  
  if(den == 1.0) {
    d = num;
    [super write: fp sf: sigfig];
  }
  else {
    d = num;
    [super write: fp sf: sigfig];
    fprintf(fp, "/");
    d = den;
    [super write: fp sf: sigfig];
  }
 
  d = tmp;
  return self;
}

-writeAccurately: (FILE *)fp {
  double tmp = d;
  
  if(den == 1.0) {
    d = num;
    [super writeAccurately: fp];
  }
  else {
    d = num;
    [super writeAccurately: fp];
    fprintf(fp, "/");
    d = den;
    [super writeAccurately: fp];
  }
 
  d = tmp;
  return self;
}

-abs {
  num = abs(num);
  return self;
}

-add: (DoubleSimple *)value {

#ifdef INTEL
  fp_pctl p = fpsetprecision(FP_PC_DBL);
#endif

 if([value isKindOf: [DoubleRational class]]) {

   // We implement a non-obvious algorithm to reduce the magnitude of the 
   // numbers involved. See Knuth (Vol II, p. 330). 
   
   double otherNum = [(DoubleRational *)value getNumerator];
   double otherDen = [(DoubleRational *)value getDenominator];
   double dOne = gcd(den, otherDen);
   
   if (dOne == 1.0) {
     double greatestCommonDivisor;
     
     num = num * otherDen + otherNum * den;
     den = den * otherDen;

     greatestCommonDivisor = gcd(num, den);
     num /= greatestCommonDivisor;
     den /= greatestCommonDivisor;

   }
   else {          // d > 1.0
     double t = num * (otherDen / dOne) + otherNum * (den / dOne);
     double dTwo = gcd(t, dOne);
     
     num = t / dTwo;
     den = (den / dOne) * (otherDen / dTwo);
     
     // and these numbers are necessarily prime to each other so there's no
     // need to reduce the fraction.
   }
   
 }
 else {  // value is not kind of rational
   
   DoubleRational *tmpRational = [DoubleRational new];
   
   [tmpRational cset: [value get]];
   
   [self add: tmpRational];
   
   [tmpRational free];
   
 }
 
#ifdef INTEL
 fpsetprecision(p);
#endif
 
 return self;
 
}

-sub: (DoubleSimple *)value {
  
#ifdef INTEL
  fp_pctl p = fpsetprecision(FP_PC_DBL);
#endif
  
  if([value isKindOf: [DoubleRational class]]) {
    
    // We implement a non-obvious algorithm to reduce the magnitude of the 
    // numbers involved. See Knuth (Vol II, p. 330). 
    
    double otherNum = [(DoubleRational *)value getNumerator];
    double otherDen = [(DoubleRational *)value getDenominator];
    double dOne = gcd(den, otherDen);
    
    if (dOne == 1.0) {
      double greatestCommonDivisor;
      
      num = num * otherDen - otherNum * den;
      den = den * otherDen;

      greatestCommonDivisor = gcd(num, den);
      num /= greatestCommonDivisor;
      den /= greatestCommonDivisor;
      
    }
    else {          // d > 1.0
      double t = num * (otherDen / dOne) - otherNum * (den / dOne);
      double dTwo = gcd(t, dOne);
      
      num = t / dTwo;
      den = (den / dOne) * (otherDen / dTwo);
      
      // and these numbers are necessarily prime to each other so there's no
      // need to reduce the fraction.
    }

  }
  else {  // value is not kind of rational

    DoubleRational *tmpRational = [DoubleRational new];

    [tmpRational cset: [value get]];

    [self sub: tmpRational];

    [tmpRational free];

  }

#ifdef INTEL
  fpsetprecision(p);
#endif

  return self;

}

-mul: (DoubleSimple *)value {

#ifdef INTEL
  fp_pctl p = fpsetprecision(FP_PC_DBL);
#endif

  if([value isKindOf: [DoubleRational class]]) {

    // We implement a non-obvious algorithm to reduce the magnitude of the 
    // numbers involved. See Knuth (Vol II, p. 330). 
    
    double otherNum = [(DoubleRational *)value getNumerator];
    double otherDen = [(DoubleRational *)value getDenominator];
    double dOne = gcd(num, otherDen);
    double dTwo = gcd(den, otherNum);
      
    num = (num / dOne) * (otherNum / dTwo);
    den = (den / dTwo) * (otherDen / dOne);

    // and these numbers are necessarily prime to each other so there's no
    // need to reduce the fraction.

  }
  else {  // value is not kind of rational

    DoubleRational *tmpRational = [DoubleRational new];

    [tmpRational cset: [value get]];

    [self mul: tmpRational];

    [tmpRational free];

  }

#ifdef INTEL
  fpsetprecision(p);
#endif

  return self;

}

-div: (DoubleSimple *)value {

#ifdef INTEL
  fp_pctl p = fpsetprecision(FP_PC_DBL);
#endif

  if([value isKindOf: [DoubleRational class]]) {

    // We implement a non-obvious algorithm to reduce the magnitude of the 
    // numbers involved. See Knuth (Vol II, p. 330). 
    
    double otherNum = [(DoubleRational *)value getNumerator];
    double otherDen = [(DoubleRational *)value getDenominator];
    double dOne = gcd(num, otherNum);
    double dTwo = gcd(den, otherDen);

    num = (num / dOne) * (otherDen / dTwo);
    den = (den / dTwo) * (otherNum / dOne);

    if(den < 0.0) {
      den = -den;
      num = -num;
    }
    
    // avoid having -0 in the numerator
    if(num == 0.0) num = 0.0;

    // and these numbers are necessarily prime to each other so there's no
    // need to reduce the fraction.

    //Finally, make sure we haven't divided by zero
    if(den == 0.0) {
      fpsetsticky(fpgetsticky() | FP_X_DZ);
      num = num < 0 ? -my_infinity : my_infinity;
    }

  }
  else {  // value is not kind of rational

    DoubleRational *tmpRational = [DoubleRational new];

    [tmpRational cset: [value get]];

    [self div: tmpRational];

    [tmpRational free];

  }

#ifdef INTEL
  fpsetprecision(p);
#endif

  return self;

}


-iadd: (int)value {

  DoubleRational *tmpRational = [DoubleRational new];
    
  tmpRational->num = (double)value;
  tmpRational->den = 1.0;

  [self add: tmpRational];
  
  [tmpRational free];
  
  return self;

}

-isub: (int)value {

  DoubleRational *tmpRational = [DoubleRational new];
    
  tmpRational->num = (double)value;
  tmpRational->den = 1.0;

  [self sub: tmpRational];
  
  [tmpRational free];
  
  return self;

}

-imul: (int)value {

  DoubleRational *tmpRational = [DoubleRational new];
    
  tmpRational->num = (double)value;
  tmpRational->den = 1.0;

  [self mul: tmpRational];
  
  [tmpRational free];
  
  return self;

}

-idiv: (int)value {

  DoubleRational *tmpRational = [DoubleRational new];

  tmpRational->num = (double)value;
  tmpRational->den = 1.0;

  [self div: tmpRational];
  
  [tmpRational free];
  
  return self;

}


-cadd: (double)value {

  DoubleRational *tmpRational = [DoubleRational new];
  [tmpRational cset: value];

  [self add: tmpRational];
  
  [tmpRational free];
  
  return self;

}

-csub: (double)value {

  DoubleRational *tmpRational = [DoubleRational new];
  [tmpRational cset: value];

  [self sub: tmpRational];
  
  [tmpRational free];
  
  return self;

}

-cmul: (double)value {

  DoubleRational *tmpRational = [DoubleRational new];
  [tmpRational cset: value];

  [self mul: tmpRational];
  
  [tmpRational free];
  
  return self;

}

-cdiv: (double)value {

  DoubleRational *tmpRational = [DoubleRational new];
  [tmpRational cset: value];

  [self div: tmpRational];
  
  [tmpRational free];
  
  return self;

}

-sadd: (char *)str {

  DoubleRational *tmpRational = [DoubleRational parseValue: str];
  
  [self add: tmpRational];
  
  [tmpRational free];

  return self;
}

-ssub: (char *)str {

  DoubleRational *tmpRational = [DoubleRational parseValue: str];
  
  [self sub: tmpRational];
  
  [tmpRational free];

  return self;
}

-smul: (char *)str {

  DoubleRational *tmpRational = [DoubleRational parseValue: str];
  
  [self mul: tmpRational];
  
  [tmpRational free];

  return self;
}

-sdiv: (char *)str {

  DoubleRational *tmpRational = [DoubleRational parseValue: str];
  
  [self div: tmpRational];
  
  [tmpRational free];

  return self;
}

-(BOOL)eq: (DoubleSimple *)value {

  if([value isKindOf: [DoubleRational class]]) {
    
    double otherNum = [(DoubleRational *)value getNumerator];
    double otherDen = [(DoubleRational *)value getDenominator];
    if(num == otherNum && den == otherDen) {
      return YES;
    }
    else {
      return NO;
    }
    
  }
  else {  // value is not kind of rational

#ifdef INTEL
    fp_pctl p = fpsetprecision(FP_PC_DBL);
#endif

    double selfValue = num / den;
    double otherValue = [value get];

#ifdef INTEL
    fpsetprecision(p);
#endif

    if(selfValue == otherValue) {
      return YES;
    }
    else {
      return NO;
    }
  }

}

-(BOOL)ne: (DoubleSimple *)value {

  if([value isKindOf: [DoubleRational class]]) {
    
    double otherNum = [(DoubleRational *)value getNumerator];
    double otherDen = [(DoubleRational *)value getDenominator];
    if(num == otherNum && den == otherDen) {
      return NO;
    }
    else {
      return YES;
    }
    
  }
  else {    // value is not kind of rational

#ifdef INTEL
    fp_pctl p = fpsetprecision(FP_PC_DBL);
#endif

    double selfValue = num / den;
    double otherValue = [value get];

#ifdef INTEL
    fpsetprecision(p);
#endif

    if(selfValue == otherValue) {
      return NO;
    }
    else {
      return YES;
    }
  }

}

-(BOOL)gt: (DoubleSimple *)value {

  if([value isKindOf: [DoubleRational class]]) {
    
    double otherNum = [(DoubleRational *)value getNumerator];
    double otherDen = [(DoubleRational *)value getDenominator];
    double dOne, dTwo;

#ifdef INTEL
    fp_pctl p = fpsetprecision(FP_PC_DBL);
#endif
    // Bear in mind that every denominator is positive

    // Quick check to avoid problems comparing zero and zero.
    if(num == otherNum && den == otherDen) {
#ifdef INTEL
      fpsetprecision(p);
#endif
      return NO;
    }

    dOne = gcd(num, otherNum);
    dTwo = gcd(den, otherDen);

    if((num / dOne) * (otherDen / dTwo) > (otherNum / dOne) * (den / dTwo)) {
#ifdef INTEL
      fpsetprecision(p);
#endif
      return YES;
    }
    else {
#ifdef INTEL
      fpsetprecision(p);
#endif
      return NO;
    }
    
  }
  else {  // value is not kind of rational

    return [self cgt: [value get]];

  }

}

-(BOOL)le: (DoubleSimple *)value {

  return ![self gt: value];

}

-(BOOL)ge: (DoubleSimple *)value {

  return ![self lt: value];
}

-(BOOL)lt: (DoubleSimple *)value {

  if([value isKindOf: [DoubleRational class]]) {
    
    double otherNum = [(DoubleRational *)value getNumerator];
    double otherDen = [(DoubleRational *)value getDenominator];
    double dOne, dTwo;

#ifdef INTEL
    fp_pctl p = fpsetprecision(FP_PC_DBL);
#endif
    // Bear in mind that every denominator is positive

    // Quick check to avoid problems comparing zero and zero.
    if(num == otherNum && den == otherDen) {
#ifdef INTEL
      fpsetprecision(p);
#endif
      return NO;
    }

    dOne = gcd(num, otherNum);
    dTwo = gcd(den, otherDen);

    if((num / dOne) * (otherDen / dTwo) < (otherNum / dOne) * (den / dTwo)) {
#ifdef INTEL
      fpsetprecision(p);
#endif
      return YES;
    }
    else {
#ifdef INTEL
      fpsetprecision(p);
#endif
      return NO;
    }
    
  }
  else {  // value is not kind of rational

    return [self clt: [value get]];

  }

}


-(BOOL)ieq: (int)value {
  BOOL answer;
  DoubleRational *tmpRational = [DoubleRational new];
    
  tmpRational->num = (double)value;
  tmpRational->den = 1.0;

  answer = [self eq: tmpRational];
  
  [tmpRational free];
  
  return answer;

}

-(BOOL)ine: (int)value {
  BOOL answer;
  DoubleRational *tmpRational = [DoubleRational new];
    
  tmpRational->num = (double)value;
  tmpRational->den = 1.0;

  answer = [self ne: tmpRational];
  
  [tmpRational free];
  
  return answer;

}

-(BOOL)igt: (int)value {
  BOOL answer;
  DoubleRational *tmpRational = [DoubleRational new];
    
  tmpRational->num = (double)value;
  tmpRational->den = 1.0;

  answer = [self gt: tmpRational];
  
  [tmpRational free];
  
  return answer;

}

-(BOOL)ile: (int)value {
  BOOL answer;
  DoubleRational *tmpRational = [DoubleRational new];
    
  tmpRational->num = (double)value;
  tmpRational->den = 1.0;

  answer = [self le: tmpRational];
  
  [tmpRational free];
  
  return answer;

}

-(BOOL)ige: (int)value {
  BOOL answer;
  DoubleRational *tmpRational = [DoubleRational new];
    
  tmpRational->num = (double)value;
  tmpRational->den = 1.0;

  answer = [self ge: tmpRational];
  
  [tmpRational free];
  
  return answer;

}

-(BOOL)ilt: (int)value {
  BOOL answer;
  DoubleRational *tmpRational = [DoubleRational new];
    
  tmpRational->num = (double)value;
  tmpRational->den = 1.0;

  answer = [self lt: tmpRational];
  
  [tmpRational free];
  
  return answer;

}

/*
There are several ways to implement the methods that compare rationals with 
floating point numbers. One possible way is to convert the floating point
number into a rational and then use the methods already implemented to compare
rationals. The following shows an example of how to do this:


-(BOOL)ceq: (double)value {
  BOOL answer;
  DoubleRational *tmpRational = [DoubleRational new];

  [tmpRational cset: value];

  answer = [self eq: tmpRational];
  
  [tmpRational free];
  
  return answer;

}

This way will very likely have floating point errors when comparing floating
point numbers with many significant bits (e.g. the closest representable fp 
numbers to simple-looking decimal numbers like 0.1, 0.2, 0.3, 0.4 ...) and
it may also give the wrong answer in many cases. For that reason I have 
decided to implement here another way of comparing rationals which may also 
give errors, but which, I think, will give the correct answer most of the 
times.

In the particular cases of checking for equality or inequality we provide a 
simple algorithm which is guaranteed to give the correct answer always, so
those two methods will not be overridden by any of the DoubleRational's 
subclasses.

*/

-(BOOL)ceq: (double)value {
  double rational;
  fp_except save = fpgetsticky();
  fp_except imprecision;
 
  fpsetsticky(0);
  rational = num/den;
  imprecision = (fpgetsticky() & FP_X_IMP);

  fpsetsticky(save); 

  if( imprecision != 0) {
    // If the rational number cannot be accurately represented in the current 
    // floating point environment, surely its value cannot be equal to any 
    // floating point number.
    return NO;
  }
  else { // rational can be exactly represented
    return (rational == value) ? YES : NO; 
  }
}

-(BOOL)cne: (double)value {
  double rational;
  fp_except save = fpgetsticky();
  fp_except imprecision;
 
  fpsetsticky(0);
  rational = num/den;
  imprecision = (fpgetsticky() & FP_X_IMP);

  fpsetsticky(save); 

  if( imprecision != 0) {
    // If the rational number cannot be accurately represented in the current 
    // floating point environment, surely its value cannot be equal to any 
    // floating point number.
    return YES;
  }
  else { // rational can be exactly represented
    return (rational != value) ? YES : NO; 
  }
}
 
-(BOOL)cgt: (double)value {
  double rational = num/den;
  return (rational > value) ? YES : NO;
}

-(BOOL)cle: (double)value {
  double rational = num/den;
  return (rational <= value) ? YES : NO;
}

-(BOOL)cge: (double)value {
  double rational = num/den;
  return (rational >= value) ? YES : NO;
}

-(BOOL)clt: (double)value {
  double rational = num/den;
  return (rational < value) ? YES : NO;
}

-(BOOL)seq: (char *)str {
  BOOL answer;
  DoubleRational *tmpRational = [DoubleRational parseValue: str];
  
  answer = [self eq: tmpRational];
  
  [tmpRational free];

  return answer;
}

-(BOOL)sne: (char *)str {
  BOOL answer;
  DoubleRational *tmpRational = [DoubleRational parseValue: str];
  
  answer = [self ne: tmpRational];
  
  [tmpRational free];

  return answer;
}

-(BOOL)sgt: (char *)str {
  BOOL answer;
  DoubleRational *tmpRational = [DoubleRational parseValue: str];
  
  answer = [self gt: tmpRational];
  
  [tmpRational free];

  return answer;
}

-(BOOL)sle: (char *)str {
  BOOL answer;
  DoubleRational *tmpRational = [DoubleRational parseValue: str];
  
  answer = [self le: tmpRational];
  
  [tmpRational free];

  return answer;
}

-(BOOL)sge: (char *)str {
  BOOL answer;
  DoubleRational *tmpRational = [DoubleRational parseValue: str];
  
  answer = [self ge: tmpRational];
  
  [tmpRational free];

  return answer;
}

-(BOOL)slt: (char *)str {
  BOOL answer;
  DoubleRational *tmpRational = [DoubleRational parseValue: str];
  
  answer = [self lt: tmpRational];
  
  [tmpRational free];

  return answer;
}


@end
